// -*- C++ -*-
/*!
 * @file  arm.cpp
 * @brief Arm
 * $Date$
 *
 * $Id$
 */

#include "arm.h"

#define TS_COUNT 1000
#define PERIOD 2000
#include "../common/timestat.h"

// Module specification
static const char* arm_spec[] =
  {
    "implementation_id", "arm",
    "type_name",         "DataFlowComponent",
    "description",       "Arm",
    "version",           "1.0",
    "vendor",            "K.Kitagaki (A shift to version 0.4.0 by S.Kurihara)",
    "category",          "arm",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };

arm::arm(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_ReferenceVelocityIn("ReferenceVelocity", m_ReferenceVelocity),
    m_CurrentPositionOut("CurrentPosition", m_CurrentPosition),
    dummy(0)
{
  // Registration: InPort/OutPort/Service
  // Set InPort buffers
  registerInPort("ReferenceVelocity", m_ReferenceVelocityIn);
  
  // Set OutPort buffer
  //  registerOutPort("CurrentPosition", m_CurrentPositionOut);
  
}

arm::~arm()
{
}


RTC::ReturnCode_t arm::onInitialize()
{
  m_palib = new Palib(ARMNO);
  m_palib->initSystem();
  m_palib->openArm();
  std::cout << "Arm Initialized" << std::endl;

  return RTC::OK;
}


RTC::ReturnCode_t arm::onFinalize()
{
  delete m_palib;
  return RTC::OK;
}


RTC::ReturnCode_t arm::onStartup(RTC::UniqueId ec_id)
{
  if (art_enter(ART_PRIO_MAX-1, ART_TASK_PERIODIC, PERIOD) == -1) {
    std::cerr << "fatal error: art_enter" << std::endl;
    return RTC::ERROR;
  }
  std::cerr << "ok: art_enter" << std::endl;
  
  int isPowerOn;
  int ret = m_palib->suspendArm(WM_WAIT);

  if (ret >= ERR_POWER_ON) {
    isPowerOn = 0;
  } else {
    isPowerOn = 1;
  }

  if (isPowerOn == 0) {
    m_palib->startArm();
    isPowerOn = 1;
  } else {
    m_palib->restartArm(WM_WAIT);
  }

  m_palib->setDIO(DP_PORT1, DC_CH4);
  usleep(1000);
  m_palib->resetDIO(DP_PORT1, DC_CH5);
  //  m_palib->setDIO(DP_PORT1, DC_CH4);
  
  m_palib->resetControl();
  
  ANGLE ang;
  
  m_palib->getAngle(ang);
  sleep(1);
  double r2d = 180. / PI;
  std::cout << 
    ang.s1 * r2d << ", " << 
    ang.s2 * r2d << ", " <<
    ang.s3 * r2d << ", " <<
    ang.e1 * r2d << ", " <<
    ang.e2 * r2d << ", " <<
    ang.w1 * r2d << ", " <<
    ang.w2 * r2d << std::endl;

  double d2r = PI / 180.;
  ang.s1 =   85.0 * d2r;
  ang.s2 =   45.0 * d2r;
  ang.s3 =    0.0 * d2r;
  ang.e1 = -110.0 * d2r;
  ang.e2 =    0.0 * d2r;
  ang.w1 =  -25.0 * d2r;

  if (ang.w2 > 0.0)
    ang.w2 = 180.0 * d2r;
  else
    ang.w2 = -180.0 * d2r;

  m_palib->execAxis((AXIS)ALLAXIS, ang, WM_WAIT);

  usleep(1);

  ang.s1 =   90.0 * d2r;
  m_palib->execAxis((AXIS)ALLAXIS, ang, WM_WAIT);  

  return RTC::OK;
}


RTC::ReturnCode_t arm::onActivated(RTC::UniqueId ec_id)
{
  m_palib->modeVelocity(VM_xyzypr, (AXIS)0);

  return RTC::OK;
}


RTC::ReturnCode_t arm::onDeactivated(RTC::UniqueId ec_id)
{
  if (art_exit() == -1) {
    std::cerr << "fatal error: art_exit" << std::endl;
    return RTC::ERROR;
  }

  m_palib->stopArm(WM_WAIT);
  //  m_palib->suspendArm(WM_WAIT);
  m_palib->setDIO(DP_PORT1, DC_CH5);
  usleep(1000);
  m_palib->resetDIO(DP_PORT1, DC_CH4);

  return RTC::OK;
}


RTC::ReturnCode_t arm::onExecute(RTC::UniqueId ec_id)
{
  static int counter;
  double t0, t1, t2;
  unsigned long long ut0, ut1, ut2;
  static double oldTime, periodTime, execTime;

  if (art_wait() == -1) {
    std::cerr << "fatal error: art_wait " << counter << std::endl;
    return RTC::ERROR;
  }

  RDTSC(ut0);
  periodTime = ut0 - oldTime;
  oldTime = ut0;
  counter++;

  float outVel[6];
  m_ReferenceVelocity = m_ReferenceVelocityIn.read();
  if (m_ReferenceVelocity.data.length() == 6) {
//     for (int i = 0; i < 6; i++) {
//       outVel[i] = m_ReferenceVelocity.data[i];
//     }
    outVel[0] = m_ReferenceVelocity.data[2];
    outVel[1] = -m_ReferenceVelocity.data[1];
    outVel[2] = m_ReferenceVelocity.data[0];
    outVel[3] = m_ReferenceVelocity.data[5];
    outVel[4] = -m_ReferenceVelocity.data[4];
    outVel[5] = m_ReferenceVelocity.data[3];
  } else {
    for (int i = 0; i < 6; i++) {
      outVel[i] = 0.0;
    }
  }

  for (int i = 0; i < 3; i++) { // $B%j%_%C%?(B
    double limitterF = 60.0; // mm/s
    double limitterR = 0.52; // rad/s
    if (outVel[i] > limitterF)
      outVel[i] = limitterF;
    else if (outVel[i] < -limitterF)
      outVel[i] = -limitterF;
    
    if (outVel[i+3] > limitterR)
      outVel[i+3] = limitterR;
    else if (outVel[i+3] < -limitterR)
      outVel[i+3] = -limitterR;
  }

  if (counter % TS_COUNT == 0) {
    for (int i = 0; i < 6; i++) {
      std::cout << outVel[i] << ", ";
    }
    std::cout << std::endl;
  }

  m_palib->orderVelocity(outVel);

  MATRIX noa;
  m_palib->getNOA(noa);

  RDTSC(ut0);
  execTime = ut0 - oldTime;
  statistics("PeriodTime ", periodTime*invClock);

  return RTC::OK;
}


extern "C"
{
 
  void armInit(RTC::Manager* manager)
  {
    RTC::Properties profile(arm_spec);
    manager->registerFactory(profile,
                             RTC::Create<arm>,
                             RTC::Delete<arm>);
  }
  
};


