// -*- C++ -*-
/*!
 * @file  controllerDamper.cpp
 * @brief controller
 * $Date$
 *
 * $Id$
 */

#include "controllerDamper.h"

#define TS_COUNT 1000
#define PERIOD 2000
#include "../common/timestat.h"

// Module specification
static const char* controllerdamper_spec[] =
  {
    "implementation_id", "controllerDamper",
    "type_name",         "DataFlowComponent",
    "description",       "controller",
    "version",           "1.0",
    "vendor",            "K.Kitagaki (A shift to version 0.4.0 by S.Kurihara)",
    "category",          "controller",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };

controllerDamper::controllerDamper(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_ReferenceForceIn("ReferenceForce", m_ReferenceForce),
    m_CurrentForceIn("CurrentForce", m_CurrentForce),
    m_DamperGainIn("DamperGain", m_DamperGain),
    m_ReferenceVelocityOut("ReferenceVelocity", m_ReferenceVelocity),
    dummy(0)
{
  // Registration: InPort/OutPort/Service
  // Set InPort buffers
  registerInPort("ReferenceForce", m_ReferenceForceIn);
  registerInPort("CurrentForce", m_CurrentForceIn);
  registerInPort("DamperGain", m_DamperGainIn);
  
  // Set OutPort buffer
  registerOutPort("ReferenceVelocity", m_ReferenceVelocityOut);
}

controllerDamper::~controllerDamper()
{
}


RTC::ReturnCode_t controllerDamper::onInitialize()
{
  m_thisCurrentForce.data.length(6);
  m_thisReferenceForce.data.length(6);
  m_thisDamperGain.data.length(6);
  for (int i = 0; i < 6; i++) {
    m_thisCurrentForce.data[i] = 0.0;
    m_thisReferenceForce.data[i] = 0.0;
    m_thisDamperGain.data[i] = 0.0;
  }  
  return RTC::OK;
}


RTC::ReturnCode_t controllerDamper::onActivated(RTC::UniqueId ec_id)
{
  std::cout << "active_entry" << std::endl;

  if (art_enter(ART_PRIO_MAX-1, ART_TASK_PERIODIC, PERIOD) == -1) {
    std::cerr << "fatal error: art_enter(active_entry)" << std::endl;
    return RTC::ERROR;
  }

  std::cout << "ok: art_enter" << std::endl;

  return RTC::OK;
}


RTC::ReturnCode_t controllerDamper::onDeactivated(RTC::UniqueId ec_id)
{
  if (art_exit() == -1) {
    std::cerr << "fatal error: art_exit" << std::endl;
    return RTC::ERROR;
  }
  std::cerr << "ok: art_enter" << std::endl;

  return RTC::OK;
}


RTC::ReturnCode_t controllerDamper::onExecute(RTC::UniqueId ec_id)
{
  static int counter;
  double t0, t1, t2;
  unsigned long long ut0, ut1, ut2;
  static double oldTime, periodTime, execTime;

  if (art_wait() == -1) {
    std::cerr << "fatal error: art_wait(active_do)" << counter << std::endl;
    return RTC::ERROR;
  }

  RDTSC(ut0);
  periodTime = ut0 - oldTime;
  oldTime = ut0;
  counter++;

  m_ReferenceForce = m_ReferenceForceIn.read();
  m_CurrentForce = m_CurrentForceIn.read();
  m_DamperGain = m_DamperGainIn.read();
  m_ReferenceVelocity.data.length(6);

  if (m_ReferenceForce.data.length() == 6) {
    for (int i = 0; i < 6; i++)
      m_thisReferenceForce.data[i] = m_ReferenceForce.data[i];
  } else {
    for (int i = 0; i < 6; i++)
      m_thisReferenceForce.data[i] = 0.0;
  }
  if (m_CurrentForce.data.length() == 6) {
    m_thisCurrentForce.data[0] = -m_CurrentForce.data[2];
    m_thisCurrentForce.data[1] = -m_CurrentForce.data[1];
    m_thisCurrentForce.data[2] =  m_CurrentForce.data[0];
    m_thisCurrentForce.data[3] = -m_CurrentForce.data[5];
    m_thisCurrentForce.data[4] = -m_CurrentForce.data[4];
    m_thisCurrentForce.data[5] =  m_CurrentForce.data[3];
  } else {
    for (int i = 0; i < 6; i++)
      m_thisCurrentForce.data[i] = 0.0;
  }
  if (m_DamperGain.data.length() == 6) {
    m_thisDamperGain.data[0] = -m_DamperGain.data[0];
    m_thisDamperGain.data[1] =  m_DamperGain.data[1];
    m_thisDamperGain.data[2] = -m_DamperGain.data[2];
    m_thisDamperGain.data[3] = -m_DamperGain.data[3];
    m_thisDamperGain.data[4] =  m_DamperGain.data[4];
    m_thisDamperGain.data[5] = -m_DamperGain.data[5];
  } else {
    for (int i = 0; i < 6; i++)
      m_thisDamperGain.data[i] = 0.0;
  }
 
   for (int i = 0; i < 6; i++) {
     m_ReferenceVelocity.data[i] = m_thisDamperGain.data[i] * (m_thisReferenceForce.data[i] - m_thisCurrentForce.data[i]);
   }

  if (counter % 1000 == 0) {
    std::cout << counter << ": ";
    std::cout << "\nm_ReferenceForce=";
    for (int i = 0; i < 6; i++) {
      std::cout << m_ReferenceForce.data[i] << ", ";
    }
    std::cout << "\nm_CurrentForce=";
    for (int i = 0; i < 6; i++) {
      std::cout << m_thisCurrentForce.data[i] << ", ";
    }
    std::cout << "\nm_DamperGain=";
    for (int i = 0; i < 6; i++) {
      std::cout << m_thisDamperGain.data[i] << ", ";
    }
    std::cout << std::endl; 
  }

  m_ReferenceVelocityOut.write(m_ReferenceVelocity);

  return RTC::OK;
}


extern "C"
{
 
  void controllerDamperInit(RTC::Manager* manager)
  {
    RTC::Properties profile(controllerdamper_spec);
    manager->registerFactory(profile,
                             RTC::Create<controllerDamper>,
                             RTC::Delete<controllerDamper>);
  }
  
};


