#include "GraphWindow.h"
#include <sys/time.h>

#ifdef ART
#include "linux/art_task.h"
#endif

//#define DEBUG

using namespace std;

GraphWindow::GraphWindow(const char* title, int width, int height)
  : win_width(width), win_height(height),
    win_pos_x(100), win_pos_y(100),
    window_name(title)
{
  erase_pixels = 5;
#ifdef CROSS
  fig[0].x = fig[0].y = 0;
  fig[1].x = fig[1].y = -1;
  fig[2].x = 2;  fig[2].y = 0;
  fig[3].x = 0;  fig[3].y = 2;
  fig[4].x = -2; fig[4].y = 0;
#else
  // circle
  fig[0].x =     fig[0].y =  0;
  fig[1].x =  0; fig[1].y = -1;
  fig[2].x =  1; fig[2].y =  1;
  fig[3].x = -1; fig[3].y =  1;
  fig[4].x = -1; fig[4].y = -1;
#endif

  timescale = 100.0; // x axis scale (sec)
  recycle = false; 
  dualGraph = false;
  forceGraph = false;
  momentGraph = false;
  fileoutput = false;
  y_offset = 0;

  // color set
  for (int i = 0; i < MAX_DATA/8; i++) {
    plotColors[0+i*8] = "red";
    plotColors[1+i*8] = "yellow";
    plotColors[2+i*8] = "green";
//   plotColors[3] = "red";
//   plotColors[4] = "yellow";
//   plotColors[5] = "green";
    plotColors[3+i*8] = "orange";
    plotColors[4+i*8] = "cyan";
    plotColors[5+i*8] = "white";
    plotColors[6+i*8] = "magenta";
    plotColors[7+i*8] = "pink";
  }

  initGraphWindow();
  initialize(); 
}

GraphWindow::~GraphWindow(void)
{
  pthread_mutex_destroy(mlmutex);
  delete mlmutex;

  for (int i = 0; i < MAX_DATA; i++)
    if (data[i].use) {
      XFreeGC(display, data[i].gc);
      data[i].pos.clear();
    }

  XFreeGC(display, gc);
  XCloseDisplay(display);
}

int GraphWindow::initGraphWindow(void)
{
  unsigned long fg,bg;
  unsigned long vmask;
  XGCValues xgcv;
  XSetWindowAttributes xswa;
  XEvent xev;
  XTextProperty win_name;
  
  if (!XInitThreads()) {
    cerr << "XInitThreads() failed." << endl;
    exit(1);
  }

  if (!(size_hints = XAllocSizeHints())) {
    cerr << "failure allocating SizeHints" << endl;
    exit(1);
  }

  char* windowName = new char[window_name.size()+1];
  strncpy(windowName, window_name.c_str(), window_name.size()+1);

  if (XStringListToTextProperty(&windowName, 1, &win_name) == 0) {
    cerr << "ERROR" << endl;
    exit(1);
  }

  size_hints->min_width  = 200;
  size_hints->min_height = 200;

  if (!(display = XOpenDisplay(NULL))) {
    cerr << "cannot open display" << endl;
    exit(1);
  }
  
  fg = setColor("white");
  bg = setColor("black");
  
  xswa.event_mask = ButtonPressMask | KeyPressMask | ExposureMask | StructureNotifyMask;
  xswa.background_pixel = bg;
  xswa.border_pixel = fg;
  
  win = XCreateWindow(display,RootWindow(display,0),
		      win_pos_x, win_pos_y,
		      win_width, win_height, 0,
		      DefaultDepth(display,DefaultScreen(display)),
		      InputOutput, DefaultVisual(display, DefaultScreen(display)),
		      CWEventMask | CWBackPixel | CWBorderPixel, &xswa);

  XSetWMName(display, win, &win_name);
  XMapWindow(display, win);
  
  for (;;) {
    XNextEvent(display, &xev);
    if (xev.type == Expose)
      break;
  }
  
  vmask = (GCBackground | GCForeground | GCLineWidth);
  xgcv.background = bg;
  xgcv.foreground = fg;
  xgcv.line_width = 1;
  
  gc  = XCreateGC (display, win, vmask, &xgcv);

  //  XSelectInput(display, win, ButtonPressMask | KeyPressMask | ExposureMask);

  // initialize mutex
  mlmutex = new pthread_mutex_t;
  if (pthread_mutex_init(mlmutex, NULL)) {
    cerr << "ERROR : initialize mutex failed." << endl;
    exit(1);
  }

  // Event_loop thread start here
  if (pthread_create(&mlthread, NULL, threadFunc, this)) {
    cerr << "ERROR : Creating thread failed." << endl;
    exit(1);
  }

  return 0;
}

void GraphWindow::initialize(void)
{
  calcGraphParameters();
  
  for (int i = 0; i < MAX_DATA; i++) {  // <- take care if you change TGT values.
    data[i].use = false;
    unsigned long vmask = (GCBackground | GCForeground | GCLineWidth);
    XGCValues xgcv;
    xgcv.background = setColor("black");
    xgcv.foreground = setColor(const_cast<char*>(plotColors[i].c_str()));
    xgcv.line_width = 2;
    data[i].gc = XCreateGC (display, win, vmask, &xgcv);
  }
  reset();
}

void GraphWindow::reset(void)
{
  orig_time = -1.;
  recycle = false;

  for (int i = 0; i < MAX_DATA; i++) {
    pthread_mutex_lock(mlmutex);
    data[i].pos.clear();
    pthread_mutex_unlock(mlmutex);
    if (data[i].use) {
      ;  // TODO : erase element here.
    }
  }
  XClearWindow(display, win);
}

void GraphWindow::setRange(double val, int f, int t)
{
  double min =  (val > 0) ? -val : val;
  double max =  (val < 0) ? -val : val;

  for (int i = f; i <= t; i++) {
    data[i].min = min;
    data[i].max = max;
    data[i].use = true;
  }
}

void GraphWindow::SetRange(int port, double min, double max)
{
    data[port].min = min;
    data[port].max = max;
    data[port].use = true;
}

void GraphWindow::SetRange2(int port, double min, double max, int graphId)
{
    data[port].min = min;
    data[port].max = max;
    data[port].use = true;
    data[port].graphId = graphId;
}

void GraphWindow::SetGraphParameters(int graphId, char* title)
{
  switch (graphId) {
  case 0:    // SetForceRange(min);
    strcpy(m_gtitle1, title); 
    forceGraph = true;
    dualGraph = momentGraph;
    calcGraphParameters();
    break;
  case 1:    // SetMomentRange(min);
    strcpy(m_gtitle2, title); 
    momentGraph = true;
    dualGraph = forceGraph;
    calcGraphParameters();
    break;
  }
}

/*
void GraphWindow::SetForceRange(double val)
{
  forceGraph = true;
  setRange(val, D1, D3);
  //  setRange(val, F1, F3);
  dualGraph = momentGraph;
  calcGraphParameters();
}

void GraphWindow::SetMomentRange(double val)
{
  momentGraph = true;
  setRange(val, D4, D6);
  //  setRange(val, M1, M3);
  dualGraph = forceGraph;
  calcGraphParameters();
}
*/

void GraphWindow::drawAxis(void)
{
  if (!forceGraph && !momentGraph)
    return;

  drawLine(x_margin, y_margin, x_margin, y_margin+graph_height, gc); // y axis
  drawLine(x_margin,             y_margin + graph_height / 2,
	   win_width - x_margin, y_margin + graph_height / 2, gc);// x axis
  
  // write graph title
  
  if (forceGraph) 
    XDrawImageString(display, win, gc, x_margin, y_margin, m_gtitle1, strlen(m_gtitle1));
  else
    XDrawImageString(display, win, gc, x_margin, y_margin, m_gtitle2, strlen(m_gtitle2));

  if (dualGraph) {
    drawLine(x_margin, y_margin + y_offset,
	     x_margin, y_margin + y_offset + graph_height , gc); // y axis
    drawLine(x_margin,             y_offset + y_margin + graph_height / 2,
	     win_width - x_margin, y_offset + y_margin + graph_height / 2, gc);// x axis
    // write second graph title
    XDrawImageString(display, win, gc, x_margin, y_margin + y_offset, m_gtitle2, strlen(m_gtitle2));
  }
}

void GraphWindow::SetDot(TGT tgt, double time, double value)
{
  if (!data[tgt].use)
    return;

  if (orig_time < 0.) {
    orig_time = time;
  }

  if (time >= orig_time + timescale) {
    orig_time += timescale;
    // TODO : erase right side of graph
  }
  
  int x, y;
  getPlotPoint(tgt, time - orig_time, value, x, y);
  point p;
  p.time = time;
  p.orig_time = orig_time;
  p.value = value;
  p.x = x;
  p.y = y;
  pthread_mutex_lock(mlmutex);
  data[tgt].pos.push_back(p);
  pthread_mutex_unlock(mlmutex);
}

void GraphWindow::SetData(double time, float value[MAX_DATA])
{
  for (int i = 0; i < MAX_DATA; i++) {
    SetDot(static_cast<TGT>(i), time, value[i]);
  }
#ifdef DEBUG
  cout << "orig_time = " << orig_time << endl;
#endif

#if 0
  XExposeEvent ev;
  ev.display = display;
  ev.window = win;
  ev.x = 1;
  ev.y = 1;
  ev.width = 1;
  ev.height = 1;
  ev.count = 0;
  ev.type = Expose;

  XSendEvent(display, win, False, ExposureMask, (XEvent*)&ev);
  XFlush(display);
#endif
}
 
void GraphWindow::getPlotPoint(int tgt, double time, double value, int& x, int& y)
{
  x = static_cast<int>((time / timescale) * (win_width - 2 * x_margin) + x_margin);
  y = static_cast<int>(graph_height + y_margin - ((value - data[tgt].min) / (data[tgt].max - data[tgt].min) * graph_height));

//   if (dualGraph && tgt >= MAX_DATA/2)
//     y += y_offset;
  y += y_offset * data[tgt].graphId;
}

void GraphWindow::ShowWindow(void)
{
  XMapWindow(display, win);
}

void GraphWindow::HideWindow(void)
{
  
}

unsigned long GraphWindow::setColor(char* colorname)
{
  XColor c0,c1;
  Colormap cmap;
  
  //  cout << colorname << endl;
  cmap = DefaultColormap(display, 0);
  assert(cmap);
  
  int ret = XAllocNamedColor(display, cmap, colorname, &c1, &c0);
  assert(ret);

  return c1.pixel;
}

void GraphWindow::drawLine(int x1, int y1, int x2, int y2, GC& gc)
{
  XDrawLine(display, win, gc, x1, y1, x2, y2);
}

void GraphWindow::Reset(void)
{
  reset();
  drawAxis();
  XFlush(display);
}

void GraphWindow::calcGraphParameters()
{
  x_margin = win_width / 20;
  y_margin = win_height / 20;
  erase_pixels = x_margin / 2;
  y_offset = dualGraph ? win_height / 2 : 0;
  graph_height = win_height - y_offset - y_margin * 2;
#ifdef DEBUG
  cout << "x_margin: " << x_margin << ", y_margin: " << y_margin
       << ", y_offset: " << y_offset << ", graph_height: " << graph_height << endl;
#endif
}

void GraphWindow::drawDot(int tgt, int x, int y)
{
// #if 0
//   fig[0].x = x;
//   fig[0].y = y;
//   XDrawPoints(display, win, data[tgt].gc, fig, POINT_NUM, CoordModePrevious);
// #endif
}

void GraphWindow::plotDot(int tgt, bool rewind)
{
  pthread_mutex_lock(mlmutex);

  int size = data[tgt].pos.size();
  point p1, p2 = data[tgt].pos[size-1];
  if (size > 1)
    p1 = data[tgt].pos[size-2];

  pthread_mutex_unlock(mlmutex);

  drawDot(tgt, p2.x, p2.y);
  if (!rewind ) {
    XDrawLine(display, win, data[tgt].gc, p1.x, p1.y, p2.x, p2.y);
  }
}

void GraphWindow::plotAll(int tgt, bool calc)
{
  point pp;
  pp.x = pp.y = win_width;

  XFlush(display); // to avoid flush in mutex time.

#ifdef DEBUG
  struct timeval tb, ta;
  gettimeofday(&tb, NULL);
#endif

  pthread_mutex_lock(mlmutex);
  for (unsigned int i = 0; i < data[tgt].pos.size(); i++) {
    point p = data[tgt].pos[i];
    if (calc)
      getPlotPoint(tgt, data[tgt].pos[i].time - data[tgt].pos[i].orig_time,
		   data[tgt].pos[i].value,
		   data[tgt].pos[i].x, data[tgt].pos[i].y);
    drawDot(tgt, data[tgt].pos[i].x, data[tgt].pos[i].y);
    if (p.x >= pp.x) {
      XDrawLine(display, win, data[tgt].gc, pp.x, pp.y, p.x, p.y);
    }
    pp.x = p.x;
    pp.y = p.y;
  }
  pthread_mutex_unlock(mlmutex);
#ifdef DEBUG
  gettimeofday(&ta, NULL);
  cout << "plotAll b" << tb.tv_sec << '.' << tb.tv_usec << endl;
  cout << "plotAll a" << ta.tv_sec << '.' << ta.tv_usec << endl;
#endif
}

bool GraphWindow::eraseArea(int tgt)
{
  static double prev_orig = orig_time; // remember the time of previous graph origin.

  pthread_mutex_lock(mlmutex);
  point pt = data[tgt].pos.front();    // erase starting point
  point pl = data[tgt].pos.back();
  pthread_mutex_unlock(mlmutex);

  if (prev_orig != orig_time) {
    XClearArea(display, win, x_margin-PLOT_WIDTH, 0, erase_pixels+PLOT_WIDTH, 0, False);
    prev_orig = orig_time;
  }

  if (pt.time < orig_time) {
    XClearArea(display, win, pl.x, 0, erase_pixels+PLOT_WIDTH, 0, False);
    return true;
  } else 
    return false;
}

bool GraphWindow::eraseData(int tgt, bool rewind) {
  static double prev_orig = -1.; // remember the time of previous graph origin.
  bool ret;

#ifdef DEBUG
  struct timeval tb, ta;
  gettimeofday(&tb, NULL);
#endif

  pthread_mutex_lock(mlmutex);
  point pt = data[tgt].pos.front();
  point pl = data[tgt].pos.back();

  // when rewind the graph data will be erased from left side in ERASE_PIXELS.
  if (prev_orig != orig_time || rewind) {
    prev_orig = orig_time;
    while (pl.x + erase_pixels >= pt.x) {
      data[tgt].pos.pop_front();
      pt = data[tgt].pos.front();
    }
    ret = true;
  } else {
    // normally, erase data within ERASE_PIXELS from latest data position.
    while (pl.x + erase_pixels >= pt.x && pl.x <= pt.x) {
      data[tgt].pos.pop_front();
      pt = data[tgt].pos.front();
    }
    ret = false;
  }
  pthread_mutex_unlock(mlmutex);
#ifdef DEBUG
  gettimeofday(&ta, NULL);
  cout << "plotAll b" << tb.tv_sec << '.' << tb.tv_usec << endl;
  cout << "plotAll a" << ta.tv_sec << '.' << ta.tv_usec << endl;
#endif
  return ret;
}

#if 0
void GraphWindow::draw(void)
#else
void GraphWindow::Draw(void)
#endif
{
  bool erased = false;
  bool rewind = false;
  for (int i = 0; i < MAX_DATA; i++) {
    if (data[i].use) {
      if (!erased)
	erased = eraseArea(i);
      if (erased)
	rewind = eraseData(i, rewind);
      plotDot(i, rewind);
#ifdef DEBUG
      cout << i << ':' << data[i].pos.size() << endl; 
#endif
    }
  }
  drawAxis();
  XFlush(display);
}

void GraphWindow::drawAll(bool calc)
{
  XClearWindow(display, win);

  for (int i = 0; i < MAX_DATA; i++)
    plotAll(i, calc);

  drawAxis();
  XFlush(display);
}

void* GraphWindow::threadFunc(void* arg) {
  static_cast<GraphWindow*>(arg)->eventProcessFunc();
  return 0;
}


void GraphWindow::eventProcessFunc(void)
{
  int i;
  bool done = false;
  bool calc = false;

  while(!done) {
    XNextEvent(display, &event);

    switch(event.type) {
    case Expose:
      if (event.xexpose.count == 0) 
#if 0
	if (event.xexpose.send_event) {
	  draw();
	} else
#endif
        if (calc) {
	  drawAll(true);
	  calc = false;
	} else
	  drawAll();      
      break;
    case ConfigureNotify:
      if (win_width == event.xconfigure.width &&
	  win_height == event.xconfigure.height)
	break;
      win_width  = event.xconfigure.width;
      win_height = event.xconfigure.height;
      if (win_width < size_hints->min_width) {
	win_width = size_hints->min_width;
      }
      if (win_height < size_hints->min_height) {
	win_height  = size_hints->min_height;
      }
      calc = true;
      calcGraphParameters();
      break;
    case KeyPress:
      char text[10];
      i = XLookupString(&event.xkey,text,10,&key,0);
      //        if (i == 1 && text[0] == 'q') {
      //    	cout << "q is pressed." << endl;
      //  	done = true;
      //        } else
      if (i == 1) {
	if (text[0] == 'q') {
#ifdef DEBUG
	  cout << "q is pressed." << endl;
	  done = true;
#endif
	}
	else if (text[0] == ' ') {
#ifdef DEBUG
	  cout << "space key is pressed." << endl;
#endif
	  fileoutput = true;
	}
#ifdef DEBUG
	else
	  cout << "other key is pressed." << endl;
#endif
      }
      break;
#ifdef DEBUG
    default:
      cout << "other event" << endl;
#endif
    }
  }

  // free everything here
}

bool GraphWindow::IsReqestFileOutput(void)
{
  bool ret = fileoutput;
  fileoutput = false;
  return ret;
}
