// -*- C++ -*-
/*!
 * @file  ftsensor.cpp
 * @brief F/T Sensor
 * $Date$
 *
 * $Id$
 */

#include "ftsensor.h"

#define TS_COUNT 1000
#define PERIOD 2000
#include "../common/timestat.h"

// Module specification
static const char* ftsensor_spec[] =
  {
    "implementation_id", "ftsensor",
    "type_name",         "DataFlowComponent",
    "description",       "F/T Sensor",
    "version",           "1.0",
    "vendor",            "K.Kitagaki (A shift to version 0.4.0 by S.Kurihara)",
    "category",          "ftsensor",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };

ftsensor::ftsensor(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_ForceOut("Force", m_Force),
    dummy(0)
{
  // Set OutPort buffer
  registerOutPort("Force", m_ForceOut);
}

ftsensor::~ftsensor()
{
}


RTC::ReturnCode_t ftsensor::onInitialize()
{
  //  cout << "FTsensor " << SENSOR_NO << endl;
  m_nittaFts = new NittaFts(SENSOR_NO);

  return RTC::OK;
}


RTC::ReturnCode_t ftsensor::onFinalize()
{
  delete m_nittaFts;
  return RTC::OK;
}


RTC::ReturnCode_t ftsensor::onActivated(RTC::UniqueId ec_id)
{
  if (art_enter(ART_PRIO_MAX-1, ART_TASK_PERIODIC, PERIOD) == -1) {
    cerr << "fatal error: art_enter(active_entry)" << endl;
    return RTC::ERROR;
  }

  cerr << "ok: art_enter" << endl;
  int fcal[6];
  m_nittaFts->calibration(fcal);
  cout << "Initialized" << endl;
  
  return RTC::OK;
}


RTC::ReturnCode_t ftsensor::onDeactivated(RTC::UniqueId ec_id)
{
  if (art_exit() == -1) {
    cerr << "fatal error: art_exit(active_exit)" << endl;
    return RTC::ERROR;
  }

  return RTC::OK;
}


RTC::ReturnCode_t ftsensor::onExecute(RTC::UniqueId ec_id)
{
  static int counter;
  double t0, t1, t2;
  unsigned long long ut0, ut1, ut2;
  static double oldTime, periodTime, execTime;
  ::DoubleVec frc;


  if (art_wait() == -1) {
    cerr << "fatal error: art_wait(active_do)" << counter << endl;
    return RTC::ERROR;
  }


  RDTSC(ut0);
  periodTime = ut0 - oldTime;
  oldTime = ut0;
  counter++;

  m_nittaFts->forceAtFsInFs(frc);

  // for Right Arm
  m_Force.data.length(6);

  for (unsigned int i = 0; i < 6; i++) {
    m_Force.data[i] = frc[i];
  }

  m_ForceOut.write(m_Force);

  RDTSC(ut0);
  execTime = ut0 - oldTime;
  statistics("PeriodTime ", periodTime*invClock);

  return RTC::OK;
}


extern "C"
{
 
  void ftsensorInit(RTC::Manager* manager)
  {
    RTC::Properties profile(ftsensor_spec);
    manager->registerFactory(profile,
                             RTC::Create<ftsensor>,
                             RTC::Delete<ftsensor>);
  }
  
};


