/*!
*   @file ExFtSensor.h
*
* Copyright 2003 Kosei Kitagaki 
*/

/*
modification history
--------------------
01b,11mar03,kk  add limit check functions
01a,10jan03,kk  first issue
*/

#ifndef __exftSensor_h
#define __exftSensor_h

//#include "SharedTable.h"
#include "EndEffector.h"
#include "FtSensor.h"
#include "ArmImpl.h"
#include "Constant.h"

// for limit checker
#define FTSLIMIT_FORCE		0x000001
#define FTSLIMIT_MOMENT		0x000002

#define FTSLIMIT_FX_UPPER	0x000100
#define FTSLIMIT_FX_LOWER	0x001000
#define FTSLIMIT_FY_UPPER	0x000200
#define FTSLIMIT_FY_LOWER	0x002000
#define FTSLIMIT_FZ_UPPER	0x000400
#define FTSLIMIT_FZ_LOWER	0x004000
#define FTSLIMIT_MX_UPPER	0x010000
#define FTSLIMIT_MX_LOWER	0x100000
#define FTSLIMIT_MY_UPPER	0x020000
#define FTSLIMIT_MY_LOWER	0x200000
#define FTSLIMIT_MZ_UPPER	0x040000
#define FTSLIMIT_MZ_LOWER	0x400000

/*!
 * 	@brief ϤΥߥå
 */ 
#define DEFAULT_LIMIT_FORCE	100.0
#define DEFAULT_LIMIT_MOMENT	10.0

class ExFtSensor {
  public:
	//! @brief 󥹥󥹤
	ExFtSensor();
	ExFtSensor(FtSensor *theFtSensor, SingleArm *theSengleArm, EndEffector *theEndEffector);

	void init(const int flag);
	void calibration(int fi[6]);
	int isCalibrated();

	void setFtsBaseCoords(const Frame &fm);
	void getFtsBaseCoords(Frame &fm);
	void setFtsCoords(const Frame &fm);
	void getFtsCoords(Frame &fm);
//	void setFtsEndCoords(const Frame &fm);
//	void getFtsEndCoords(Frame &fm);

	void setFilter(const int filter);
	int getFilter(void);

	void setGravity(const Pos &g);
	void getGravity(Pos &g);

/*
  There are several force/torque data.
  at measurement coordinates origin and
  in coordinates.

  At-In  FSC AWC  FCC  WC   TC
  FSCO   -/g  x    x    x    x  
  AWCO    x  -/g  -/g  o/g  o/g
  FCCO    x   x   -/g  o/g  o/g
  WCO     x   x    x    x    x
  TCO     x   x   o/g  o/g  o/g

  x; not supported.

  -/g; not gravity compensated force/torque defined in FtSensor class and 
       this class has adapters.
       gravity compensated force/torque defined in this class.

  o/g; both of force/torque defined in this class.

  FSCO; F/T Sensor Coordinates Origin
  AWCO; Arm Wrist Coordinates Origin
  FCCO; Force Control Coordinates Origin
  WCO;  World Coordinates Origin
  TCO;  Tool Coordinates Origin

*/
	void forceAtFsInFs(DoubleVec& f);
	void forceAtAwInAw(DoubleVec& f);
	void forceAtAwInFc(DoubleVec& f);
	void forceAtFcInFc(DoubleVec& f);

	void forceAtAwInWc(DoubleVec& f);
	void forceAtAwInTc(DoubleVec& f);
	void forceAtFcInWc(DoubleVec& f);
	void forceAtFcInTc(DoubleVec& f);
	void forceAtTcInFc(DoubleVec& f);
	void forceAtTcInWc(DoubleVec& f);
	void forceAtTcInTc(DoubleVec& f);

	void getGcForceAtAwInAw(const Frame& wristConf, DoubleVec& f);
	void gcCalibration(void);
	BOOL isGcCalibrated(void);

	void gcForceAtFsInFs(DoubleVec& f);
	void gcForceAtAwInAw(DoubleVec& f);
	void gcForceAtAwInFc(DoubleVec& f);
	void gcForceAtFcInFc(DoubleVec& f);
	void gcForceAtAwInWc(DoubleVec& f);
	void gcForceAtAwInTc(DoubleVec& f);
	void gcForceAtFcInWc(DoubleVec& f);
	void gcForceAtFcInTc(DoubleVec& f);
	void gcForceAtTcInFc(DoubleVec& f);
	void gcForceAtTcInWc(DoubleVec& f);
	void gcForceAtTcInTc(DoubleVec& f);

	void transAtAwInAw2AtAwInWc(DoubleVec& frc, DoubleVec& f);
	void transAtAwInAw2AtAwInTc(DoubleVec& frc, DoubleVec& f);
	void transAtFcInFc2AtFcInWc(DoubleVec& frc, DoubleVec& f);
	void transAtFcInFc2AtFcInTc(DoubleVec& frc, DoubleVec& f);
	void transAtTcInTc2AtTcInFc(DoubleVec& frc, DoubleVec& f);
	void transAtTcInTc2AtTcInWc(DoubleVec& frc, DoubleVec& f);
	void transAtAwInAw2AtTcInTc(DoubleVec& frc, DoubleVec& f);

// Limit checker
	void initAbsLimit(void);
	void setAbsLimit(const float limitFrc, const float limitMmt);
	void getAbsLimit(float* limitFrc, float* limitMmt);
	BOOL checkAbsLimit(const DoubleVec& frc, int *stat);

	void initAllLimit();
	void setAllLimit(const DoubleVec& limitFrcLower, const DoubleVec& limitFrcUpper);
	void getAllLimit(DoubleVec& limitFrcLower, DoubleVec& limitFrcUpper);
	BOOL checkAllLimit(const DoubleVec& frc, int *stat);

  private:
	FtSensor* myFtSensor;
	SingleArm* mySingleArm;
	EndEffector* myEndEffector;
	Pos gravityInWorld;
	Pos mGravityInWorld;
	DoubleVec offsetGravityAtAwInAw;
	int gcCalibrated;

//  EndEffector parameters
	SCALAR myMass;
	Pos myCenterOfGravity;
	Pos myCenterOfGravityMeter;
	Matrix3x3 myMomentOfInertia;

// for limit check
	SCALAR myAbsLimitFrc, myAbsLimitMmt;
	DoubleVec myAllLimitFrcLower, myAllLimitFrcUpper;
};

#endif

